import { Button, Grid, Paper } from "@mui/material";
import createStyles from '@mui/styles/createStyles';
import makeStyles from '@mui/styles/makeStyles';
import { Field, Form, Formik, FormikHelpers, FormikProps } from "formik";
import { FC, useCallback, useMemo, useRef, useState } from "react";
import { Editor } from "../commons/editor/vditor";
import * as Yup from "yup";
import {
  Article,
  CreateArticleInput,
  CreateDraftArticleInput,
  DraftArticle,
  UpdateArticleInput,
} from "../generated/graphql";
import { DateTimePicker } from "formik-material-ui-pickers";
import { gql, useMutation } from "@apollo/client";
import { useRouter } from "@curi/react-universal";
import * as R from "ramda";

const CREATE_ARTICLE = gql`
  mutation createArticle($createArticleInput: CreateArticleInput!) {
    createArticle(createArticleInput: $createArticleInput) {
      id
      title
      content
      publishedAt
    }
  }
`;

const UPDATE_ARTICLE = gql`
  mutation updateArticle($updateArticleInput: UpdateArticleInput!) {
    updateArticle(updateArticleInput: $updateArticleInput) {
      id
      title
      content
      publishedAt
    }
  }
`;

const useStyles = makeStyles((theme) =>
  createStyles({
    form: {
      padding: "15px 30px",
    },
    editor: {
      height: "700px",
    },
  })
);

const fields = ["title", "content", "publishedAt", "tags"] as const;
type Values = Pick<Article, typeof fields[number]> &
  Partial<Pick<Article, "id" | "key">>;
interface Props {
  article?: Values;
}

export const ArticleEditor: FC<Props> = ({ article }) => {
  const classes = useStyles();
  const router = useRouter();
  const formik = useRef<FormikProps<Values>>(null);

  const validationSchema = Yup.object({
    content: Yup.string()
      .max(65535, "文章内容不得超过 65535 个字符")
      .required("文章内容不得为空"),
    publishedAt: Yup.date(),
  });

  const [createArticle] = useMutation<
    { createArticle: Article },
    {
      createArticleInput: CreateArticleInput;
    }
  >(CREATE_ARTICLE, {
    update(cache, { data }) {
      cache.modify({
        fields: {
          articles(existingArticles = []) {
            const newArticleRef = cache.writeFragment({
              data: data!.createArticle,
              fragment: gql`
                fragment NewArticle on Article {
                  id
                  title
                  content
                  publishedAt
                }
              `,
            });
            return [newArticleRef, ...existingArticles];
          },
        },
      });
    },
  });
  const [updateArticle] = useMutation<{
    updateArticleInput: UpdateArticleInput;
  }>(UPDATE_ARTICLE);

  const [putDraft] = useMutation<
    {
      draft: DraftArticle;
    },
    {
      draft: CreateDraftArticleInput;
    }
  >(gql`
    mutation ($draft: CreateDraftArticleInput!) {
      draft: putDraftForArticle(draft: $draft) {
        id
        createdAt
      }
    }
  `);

  const tempKey = useMemo(() => {
    if (article?.key) {
      return article.key;
    } else {
      return new Date().valueOf().toString();
    }
  }, [article]);

  const submitForm = async (
    values: Values,
    { setSubmitting }: FormikHelpers<Values>
  ) => {
    if ("id" in article!) {
      const title = /# (.+)$/m.exec(values.content ?? "")?.[1];
      if (!title) {
        console.log("no title");
        setSubmitting(false);
        return;
      }
      await updateArticle({
        variables: {
          updateArticleInput: {
            ...values,
            title,
          },
        },
      })
        .then(() => {
          router.navigate({
            url: router.url({ name: "articles" }),
            method: "replace",
          });
        })
        .finally(() => {
          setSubmitting(false);
        });
    } else {
      const title = /# (.+)$/m.exec(values.content ?? "")?.[1];
      if (!title) {
        console.log("no title");
        setSubmitting(false);
        return;
      }
      await createArticle({
        variables: {
          createArticleInput: {
            ...(values as CreateArticleInput),
            title,
          },
        },
      })
        .then(() => {
          router.navigate({
            url: router.url({ name: "articles" }),
            method: "replace",
          });
        })
        .finally(() => {
          setSubmitting(false);
        });
    }
    setSubmitting(false);
  };

  const [drafts, setDrafts] = useState<DraftArticle[]>([]);
  const handleSave = useCallback(
    async (automatically = true) => {
      const { data } = await putDraft({
        variables: {
          draft: {
            key: tempKey,
            payload: formik.current?.values,
            automatically,
          },
        },
      });
      if (!data?.draft) {
        return;
      }
      setDrafts((drafts) => {
        const index = drafts.findIndex((it) => it.id === data.draft.id);
        if (index === -1) {
          return [data.draft, ...drafts];
        } else {
          const arr = [...drafts];
          arr[index] = data.draft;
          return arr;
        }
      });
    },
    [tempKey, putDraft, formik, drafts]
  );

  return (
    <Paper>
      <Formik
        initialValues={{
          ...R.pick(fields, article!),
        }}
        innerRef={formik}
        validationSchema={validationSchema}
        autoComplete="off"
        onSubmit={submitForm}
      >
        {({ submitForm, isSubmitting }) => (
          <Form className={classes.form}>
            <Grid container spacing={3}>
              <Grid item xs={12}>
                <Editor
                  className={classes.editor}
                  name="content"
                  label="Content"
                />
              </Grid>
              <Grid item xs={12}>
                <Field
                  ampm={false}
                  component={DateTimePicker}
                  label="Published At"
                  name="publishedAt"
                  format="yyyy-MM-dd HH:mm:ss"
                />
              </Grid>
              <Grid item xs={12}>
                <Button
                  variant="contained"
                  color="primary"
                  disabled={isSubmitting}
                  onClick={() => handleSave(false)}
                >
                  Save
                </Button>
                <Button
                  style={{ marginLeft: "10px" }}
                  variant="contained"
                  color="secondary"
                  disabled={isSubmitting}
                  onClick={submitForm}
                >
                  Publish
                </Button>
              </Grid>
            </Grid>
          </Form>
        )}
      </Formik>
    </Paper>
  );
};

ArticleEditor.defaultProps = {
  article: {
    title: "",
    content: "",
    publishedAt: new Date(),
    tags: [],
  } as Values,
};
